# -*- coding: utf-8 -*-

import importlib
from os import environ as env
from os import getpid
from unittest.mock import patch

import pytest

import expandvars


def test_unset_var_substring():
    assert expandvars.expandvars("${SOME_UNKNOWN_VAR:0:10}") == ""


def test_unset_var_length():
    assert expandvars.expandvars("${#SOME_UNKNOWN_VAR}") == "0"


#
@patch.dict(env, {}, clear=True)
def test_expandvars_constant():
    importlib.reload(expandvars)

    assert expandvars.expandvars("FOO") == "FOO"
    assert expandvars.expandvars("$") == "$"
    assert expandvars.expandvars("BAR$") == "BAR$"


@patch.dict(env, {}, clear=True)
def test_expandvars_empty():
    importlib.reload(expandvars)

    assert expandvars.expandvars("") == ""
    assert expandvars.expandvars("$FOO") == ""


@patch.dict(env, {"FOO": "bar"}, clear=True)
def test_expandvars_simple():
    importlib.reload(expandvars)

    assert expandvars.expandvars("$FOO") == "bar"
    assert expandvars.expandvars("${FOO}") == "bar"


@patch.dict(env, {"FOO": "bar"}, clear=True)
def test_expandvars_from_file():
    importlib.reload(expandvars)

    with open("tests/data/foo.txt") as f:
        assert expandvars.expandvars(f) == "bar:bar\n"


@patch.dict(env, {"FOO": "bar", "BIZ": "buz"}, clear=True)
def test_expandvars_combo():
    importlib.reload(expandvars)

    assert expandvars.expandvars("${FOO}:$BIZ") == "bar:buz"
    assert expandvars.expandvars("$FOO$BIZ") == "barbuz"
    assert expandvars.expandvars("${FOO}$BIZ") == "barbuz"
    assert expandvars.expandvars("$FOO${BIZ}") == "barbuz"
    assert expandvars.expandvars("$FOO-$BIZ") == "bar-buz"
    assert expandvars.expandvars("boo$BIZ") == "boobuz"
    assert expandvars.expandvars("boo${BIZ}") == "boobuz"


@patch.dict(env, {}, clear=True)
def test_expandvars_pid():
    importlib.reload(expandvars)

    assert expandvars.expandvars("$$") == str(getpid())
    assert expandvars.expandvars("PID( ${$} )") == "PID( {0} )".format(getpid())


@patch.dict(env, {"ALTERNATE": "Alternate", "EMPTY": ""}, clear=True)
def test_expandvars_get_default():
    importlib.reload(expandvars)

    assert expandvars.expandvars("${FOO-default}") == "default"
    assert expandvars.expandvars("${FOO:-default}") == "default"
    assert expandvars.expandvars("${EMPTY:-default}") == "default"
    assert expandvars.expandvars("${FOO:-}") == ""
    assert expandvars.expandvars("${FOO:-foo}:${FOO-bar}") == "foo:bar"
    assert expandvars.expandvars("${FOO:-$ALTERNATE}") == "Alternate"
    assert expandvars.expandvars("${UNSET:-\\$foo}-\\$foo") == "$foo-$foo"


@patch.dict(env, {"EMPTY": ""}, clear=True)
def test_expandvars_update_default():
    importlib.reload(expandvars)

    assert expandvars.expandvars("${FOO:=}") == ""

    assert expandvars.expandvars("${FOO=}") == ""
    assert expandvars.expandvars("${EMPTY:=}") == ""

    del env["FOO"]
    del env["EMPTY"]

    assert expandvars.expandvars("${FOO:=default}") == "default"
    assert expandvars.expandvars("${FOO=default}") == "default"
    assert expandvars.expandvars("${EMPTY:=default}") == "default"
    assert env.get("FOO") == "default"
    assert expandvars.expandvars("${FOO:=ignoreme}") == "default"
    assert expandvars.expandvars("${EMPTY:=ignoreme}") == "default"
    assert expandvars.expandvars("${FOO=ignoreme}:bar") == "default:bar"


@patch.dict(env, {"FOO": "bar", "BUZ": "bar", "EMPTY": ""}, clear=True)
def test_expandvars_substitute():
    importlib.reload(expandvars)

    assert expandvars.expandvars("${FOO:+foo}") == "foo"
    assert expandvars.expandvars("${FOO+foo}") == "foo"
    assert expandvars.expandvars("${BAR:+foo}") == ""
    assert expandvars.expandvars("${BAR+foo}") == ""
    assert expandvars.expandvars("${EMPTY:+foo}") == ""
    assert expandvars.expandvars("${BAR:+}") == ""
    assert expandvars.expandvars("${BAR+}") == ""
    assert expandvars.expandvars("${BUZ:+foo}") == "foo"
    assert expandvars.expandvars("${BUZ+foo}:bar") == "foo:bar"
    assert expandvars.expandvars("${FOO:+${FOO};}") == "bar;"
    assert expandvars.expandvars("${BAR:+${BAR};}") == ""
    assert expandvars.expandvars("${BAR:+${EMPTY};}") == ""
    assert expandvars.expandvars("${FOO:+\\$foo}-\\$foo") == "$foo-$foo"


@patch.dict(env, {"FOO": "damnbigfoobar", "THREE": "3"}, clear=True)
def test_offset():
    importlib.reload(expandvars)

    assert expandvars.expandvars("${FOO:3}") == "nbigfoobar"
    assert expandvars.expandvars("${FOO:$THREE}") == "nbigfoobar"
    assert expandvars.expandvars("${FOO:${THREE}}") == "nbigfoobar"
    assert expandvars.expandvars("${FOO: 4 }") == "bigfoobar"
    assert expandvars.expandvars("${FOO:30}") == ""
    assert expandvars.expandvars("${FOO:0}") == "damnbigfoobar"
    assert expandvars.expandvars("${FOO: }") == "damnbigfoobar"


@patch.dict(env, {"FOO": "damnbigfoobar", "TWO": "2"}, clear=True)
def test_offset_length():
    importlib.reload(expandvars)

    assert expandvars.expandvars("${FOO: : }") == ""
    assert expandvars.expandvars("${FOO:-3}:bar") == "damnbigfoobar:bar"
    assert expandvars.expandvars("${FOO::}") == ""
    assert expandvars.expandvars("${FOO::aaa}") == ""
    assert expandvars.expandvars("${FOO: :2}") == "da"
    assert expandvars.expandvars("${FOO: :$TWO}") == "da"
    assert expandvars.expandvars("${FOO: :${TWO}}") == "da"
    assert expandvars.expandvars("${FOO:aaa:2}") == "da"
    assert expandvars.expandvars("${FOO:aaa:$TWO}") == "da"
    assert expandvars.expandvars("${FOO:aaa:${TWO}}") == "da"
    assert expandvars.expandvars("${FOO:4:3}") == "big"
    assert expandvars.expandvars("${FOO: 7:6 }") == "foobar"
    assert expandvars.expandvars("${FOO:7: 100 }") == "foobar"
    assert expandvars.expandvars("${FOO:0:100}") == "damnbigfoobar"
    assert expandvars.expandvars("${FOO:70:10}") == ""
    assert expandvars.expandvars("${FOO:1:0}") == ""
    assert expandvars.expandvars("${FOO:0:}") == ""
    assert expandvars.expandvars("${FOO::}") == ""
    assert expandvars.expandvars("${FOO::5}") == "damnb"
    assert expandvars.expandvars("${FOO:-3:1}:bar") == "damnbigfoobar:bar"


#
@patch.dict(env, {"FOO": "X", "X": "foo"}, clear=True)
def test_expandvars_indirection():
    importlib.reload(expandvars)

    assert expandvars.expandvars("${!FOO}:${FOO}") == "foo:X"
    assert expandvars.expandvars("${!FOO-default}") == "foo"
    assert expandvars.expandvars("${!X-default}") == "default"


@patch.dict(env, {"THREE": "abc", "FOUR": "abcd", "SIX": "abcdef"})
def test_length():
    importlib.reload(expandvars)

    assert expandvars.expandvars("${#THREE}") == "3"
    assert expandvars.expandvars("${#THREE}${#FOUR}") == "34"
    assert expandvars.expandvars("foo${#SIX}$SIX") == "foo6abcdef"


@patch.dict(env, {"FOO": "foo", "BAR": "bar"}, clear=True)
def test_escape():
    importlib.reload(expandvars)

    assert expandvars.expandvars("\\$FOO\\$BAR") == "$FOO$BAR"
    assert expandvars.expandvars("\\\\$FOO") == "\\foo"
    assert expandvars.expandvars("$FOO\\$BAR") == "foo$BAR"
    assert expandvars.expandvars("\\$FOO$BAR") == "$FOObar"
    assert expandvars.expandvars("$FOO\\\\\\$BAR") == ("foo\\$BAR")
    assert expandvars.expandvars("$FOO\\$") == "foo$"
    assert expandvars.expandvars("$\\FOO") == "$\\FOO"
    assert expandvars.expandvars("$\\$FOO") == "$$FOO"
    assert expandvars.expandvars("\\$FOO") == "$FOO"
    assert (
        expandvars.expandvars("D:\\\\\\some\\windows\\path")
        == "D:\\\\some\\windows\\path"
    )


#
@patch.dict(env, {}, clear=True)
def test_corner_cases():
    importlib.reload(expandvars)

    assert expandvars.expandvars("${FOO:-{}}{}{}{}{{}}") == "{}{}{}{}{{}}"
    assert expandvars.expandvars("${FOO-{}}{}{}{}{{}}") == "{}{}{}{}{{}}"


@patch.dict(env, {}, clear=True)
def test_strict_parsing():
    importlib.reload(expandvars)

    with pytest.raises(
        expandvars.ExpandvarsException, match="FOO: parameter null or not set"
    ) as e:
        expandvars.expandvars("${FOO:?}")
    assert isinstance(e.value, expandvars.ParameterNullOrNotSet)

    with pytest.raises(
        expandvars.ExpandvarsException, match="FOO: parameter null or not set"
    ) as e:
        expandvars.expandvars("${FOO?}")
    assert isinstance(e.value, expandvars.ParameterNullOrNotSet)

    with pytest.raises(expandvars.ExpandvarsException, match="FOO: custom error") as e:
        expandvars.expandvars("${FOO:?custom error}")
    assert isinstance(e.value, expandvars.ParameterNullOrNotSet)

    with pytest.raises(expandvars.ExpandvarsException, match="FOO: custom error") as e:
        expandvars.expandvars("${FOO?custom error}")
    assert isinstance(e.value, expandvars.ParameterNullOrNotSet)

    env.update({"FOO": "foo"})

    assert expandvars.expandvars("${FOO:?custom err}") == "foo"
    assert expandvars.expandvars("${FOO?custom err}:bar") == "foo:bar"


@patch.dict(env, {"FOO": "foo"}, clear=True)
def test_missing_escapped_character():
    importlib.reload(expandvars)

    with pytest.raises(expandvars.ExpandvarsException) as e:
        expandvars.expandvars("$FOO\\")

    assert str(e.value) == "$FOO\\: missing escaped character"
    assert isinstance(e.value, expandvars.MissingEscapedChar)


@patch.dict(env, {"FOO": "damnbigfoobar"}, clear=True)
def test_invalid_length_err():
    importlib.reload(expandvars)

    with pytest.raises(
        expandvars.ExpandvarsException, match="FOO: -3: substring expression < 0"
    ) as e:
        expandvars.expandvars("${FOO:1:-3}")
    assert isinstance(e.value, expandvars.NegativeSubStringExpression)


@patch.dict(env, {"FOO": "damnbigfoobar"}, clear=True)
def test_bad_substitution_err():
    importlib.reload(expandvars)

    with pytest.raises(expandvars.ExpandvarsException) as e:
        expandvars.expandvars("${FOO:}")
    assert str(e.value) == "${FOO:}: bad substitution"
    assert isinstance(e.value, expandvars.BadSubstitution)

    with pytest.raises(expandvars.ExpandvarsException) as e:
        expandvars.expandvars("${}")
    assert str(e.value) == "${}: bad substitution"
    assert isinstance(e.value, expandvars.BadSubstitution)

    with pytest.raises(expandvars.ExpandvarsException) as e:
        expandvars.expandvars("${FOO:1:2:3}")
    assert str(e.value) == "${FOO:1:2:3}: bad substitution"
    assert isinstance(e.value, expandvars.BadSubstitution)

    with pytest.raises(expandvars.ExpandvarsException) as e:
        expandvars.expandvars("${#FOO:-default}")
    assert str(e.value) == "${#FOO:-default}: bad substitution"
    assert isinstance(e.value, expandvars.BadSubstitution)

    with pytest.raises(expandvars.ExpandvarsException) as e:
        expandvars.expandvars("${#FOO:1}")
    assert str(e.value) == "${#FOO:1}: bad substitution"
    assert isinstance(e.value, expandvars.BadSubstitution)


@patch.dict(env, {}, clear=True)
def test_invalid_indirect_expansion():
    importlib.reload(expandvars)

    with pytest.raises(
        expandvars.ExpandvarsException, match="FOO: invalid indirect expansion"
    ) as e:
        expandvars.expandvars("${!FOO}")

    assert isinstance(e.value, expandvars.InvalidIndirectExpansion)


@patch.dict(env, {"FOO": "damnbigfoobar"}, clear=True)
def test_brace_never_closed_err():
    importlib.reload(expandvars)

    with pytest.raises(expandvars.ExpandvarsException) as e:
        expandvars.expandvars("${FOO:")
    assert str(e.value) == "${FOO:: missing '}'"
    assert isinstance(e.value, expandvars.MissingClosingBrace)

    with pytest.raises(expandvars.ExpandvarsException) as e:
        expandvars.expandvars("${FOO}${BAR")
    assert str(e.value) == "${FOO}${BAR: missing '}'"
    assert isinstance(e.value, expandvars.MissingClosingBrace)

    with pytest.raises(expandvars.ExpandvarsException) as e:
        expandvars.expandvars("${FOO?")
    assert str(e.value) == "${FOO?: missing '}'"
    assert isinstance(e.value, expandvars.ExpandvarsException)

    with pytest.raises(expandvars.ExpandvarsException) as e:
        expandvars.expandvars("${FOO:1")
    assert str(e.value) == "${FOO:1: missing '}'"
    assert isinstance(e.value, expandvars.MissingClosingBrace)

    with pytest.raises(expandvars.ExpandvarsException) as e:
        expandvars.expandvars("${FOO:1:2")
    assert str(e.value) == "${FOO:1:2: missing '}'"
    assert isinstance(e.value, expandvars.MissingClosingBrace)

    with pytest.raises(expandvars.ExpandvarsException) as e:
        expandvars.expandvars("${FOO+")
    assert str(e.value) == "${FOO+: missing '}'"
    assert isinstance(e.value, expandvars.MissingClosingBrace)

    with pytest.raises(expandvars.ExpandvarsException) as e:
        expandvars.expandvars("${FOO-")
    assert str(e.value) == "${FOO-: missing '}'"
    assert isinstance(e.value, expandvars.MissingClosingBrace)

    with pytest.raises(expandvars.ExpandvarsException) as e:
        expandvars.expandvars("${FOO-{{}")
    assert str(e.value) == "${FOO-{{}: missing '}'"
    assert isinstance(e.value, expandvars.MissingClosingBrace)


@patch.dict(env, {"FOO": "damnbigfoobar"}, clear=True)
def test_invalid_operand_err():
    importlib.reload(expandvars)

    oprnds = "@#$%^&*()'\""

    for o in oprnds:
        with pytest.raises(expandvars.ExpandvarsException) as e:
            print(o)
            expandvars.expandvars("${{FOO:0:{0}}}".format(o))
        assert str(e.value) == ("FOO: operand expected (error token is {0})").format(
            repr(o)
        )
        assert isinstance(e.value, expandvars.OperandExpected)

        with pytest.raises(expandvars.ExpandvarsException) as e:
            expandvars.expandvars("${{FOO:{0}:{0}}}".format(o))
        assert str(e.value) == ("FOO: operand expected (error token is {0})").format(
            repr(o)
        )
        assert isinstance(e.value, expandvars.OperandExpected)


@pytest.mark.parametrize("var_symbol", ["%", "&", "£", "="])
def test_expand_var_symbol(var_symbol):
    importlib.reload(expandvars)

    assert (
        expandvars.expand(
            var_symbol + "{FOO}", environ={"FOO": "test"}, var_symbol=var_symbol
        )
        == "test"
    )
    assert (
        expandvars.expand(var_symbol + "FOO", environ={}, var_symbol=var_symbol) == ""
    )
    assert (
        expandvars.expand(
            var_symbol + "{FOO:-default_value}", environ={}, var_symbol=var_symbol
        )
        == "default_value"
    )
    with pytest.raises(expandvars.ParameterNullOrNotSet):
        expandvars.expand(var_symbol + "{FOO:?}", environ={}, var_symbol=var_symbol)

    assert (
        expandvars.expand(
            var_symbol + "{FOO},$HOME", environ={"FOO": "test"}, var_symbol=var_symbol
        )
        == "test,$HOME"
    )


@patch.dict(env, {"FOO": "bar", "BIZ": "buz"}, clear=True)
def test_expandvars_require_suffix():
    importlib.reload(expandvars)

    assert expandvars.expand("${FOO}:$BIZ", surrounded_vars_only=True) == "bar:$BIZ"
    assert expandvars.expand("$FOO$BIZ", surrounded_vars_only=True) == "$FOO$BIZ"
    assert expandvars.expand("${FOO}$BIZ", surrounded_vars_only=True) == "bar$BIZ"
    assert expandvars.expand("$FOO${BIZ}", surrounded_vars_only=True) == "$FOObuz"
    assert expandvars.expand("$FOO-$BIZ", surrounded_vars_only=True) == "$FOO-$BIZ"
    assert expandvars.expand("boo$BIZ", surrounded_vars_only=True) == "boo$BIZ"
    assert expandvars.expand("boo${BIZ}", surrounded_vars_only=True) == "boobuz"


@patch.dict(env, {"FOO": "bar", "BIZ": "buz"}, clear=True)
def test_expandvars_disable_escape():
    importlib.reload(expandvars)

    assert (
        expandvars.expand("\\foo\\", surrounded_vars_only=True, escape_char=None)
        == "\\foo\\"
    )



# -*- coding: utf-8 -*-

import os
from io import TextIOWrapper

__author__ = "Arijit Basu"
__email__ = "sayanarijit@gmail.com"
__homepage__ = "https://github.com/sayanarijit/expandvars"
__description__ = "Expand system variables Unix style"
__version__ = "v1.1.2"
__license__ = "MIT"
__all__ = [
    "BadSubstitution",
    "ExpandvarsException",
    "MissingClosingBrace",
    "MissingEscapedChar",
    "NegativeSubStringExpression",
    "OperandExpected",
    "ParameterNullOrNotSet",
    "UnboundVariable",
    "expand",
    "expandvars",
]


ESCAPE_CHAR = "\\"
VAR_SYMBOL = "$"


class ExpandvarsException(Exception):
    """The base exception for all the handleable exceptions."""

    pass


class MissingClosingBrace(ExpandvarsException, SyntaxError):
    def __init__(self, param):
        super().__init__("{0}: missing '}}'".format(param))


class MissingEscapedChar(ExpandvarsException, SyntaxError):
    def __init__(self, param):
        super().__init__("{0}: missing escaped character".format(param))


class OperandExpected(ExpandvarsException, SyntaxError):
    def __init__(self, param, operand):
        super().__init__(
            "{0}: operand expected (error token is {1})".format(param, repr(operand))
        )


class NegativeSubStringExpression(ExpandvarsException, IndexError):
    def __init__(self, param, expr):
        super().__init__("{0}: {1}: substring expression < 0".format(param, expr))


class BadSubstitution(ExpandvarsException, SyntaxError):
    def __init__(self, param):
        super().__init__("{0}: bad substitution".format(param))


class ParameterNullOrNotSet(ExpandvarsException, KeyError):
    def __init__(self, param, msg=None):
        if msg is None:
            msg = "parameter null or not set"
        super().__init__("{0}: {1}".format(param, msg))


class UnboundVariable(ExpandvarsException, KeyError):
    def __init__(self, param):
        super().__init__("{0}: unbound variable".format(param))


class InvalidIndirectExpansion(ExpandvarsException, KeyError):
    def __init__(self, param):
        super().__init__("{0}: invalid indirect expansion".format(param))


def getenv(var, indirect, environ, var_symbol=VAR_SYMBOL):
    """Get value from environment variable.

    When indirect is True, it will use the value of the resolved variable as
    the name of the final variable.
    """

    if var == var_symbol:
        return str(os.getpid())

    val = environ.get(var)

    if indirect:
        if val is None:
            raise InvalidIndirectExpansion(var)
        else:
            val = getenv(
                val,
                indirect=False,
                environ=environ,
                var_symbol=var_symbol,
            )

    return val


def expand(
    vars_,
    nounset=False,
    environ=os.environ,
    var_symbol=VAR_SYMBOL,
    surrounded_vars_only=False,
    escape_char=ESCAPE_CHAR,
):
    """Expand variables Unix style.

    Params:
        vars_ (str):  Variables to expand.
        nounset (bool): If True, enables strict parsing (similar to set -u / set -o nounset in bash).
        environ (Mapping): Elements to consider during variable expansion. Defaults to os.environ
        var_symbol (str): Character used to identify a variable. Defaults to $

    Returns:
        str: Expanded values.

    Example usage: ::

        from expandvars import expand

        print(expand("%PATH:$HOME/bin:%{SOME_UNDEFINED_PATH:-/default/path}", environ={"PATH": "/example"}, var_symbol="%"))
        # /example:$HOME/bin:/default/path

        # Or
        with open(somefile) as f:
            print(expand(f))
    """
    if isinstance(vars_, TextIOWrapper):
        # This is a file. Read it.
        vars_ = vars_.read()

    if len(vars_) == 0:
        return ""

    buff = []

    vars_iter = _PeekableIterator(vars_)
    try:
        for c in vars_iter:
            if escape_char and c == escape_char:
                next_ = vars_iter.peek()
                if next_ == var_symbol or next_ == escape_char:
                    buff.append(next(vars_iter))
                elif next_ == _PeekableIterator.NOTHING:
                    raise MissingEscapedChar(c)
                else:
                    buff.append(c)
                    buff.append(next(vars_iter))
            elif c == var_symbol:
                next_ = vars_iter.peek()
                if next_ == _PeekableIterator.NOTHING:
                    buff.append(c)
                elif surrounded_vars_only and next_ != "{":
                    buff.append(c)
                elif _valid_char(next_) or next_ == "{" or next_ == var_symbol:
                    val = _expand_var(
                        vars_iter,
                        nounset=nounset,
                        environ=environ,
                        var_symbol=var_symbol,
                    )
                    buff.append(val)
                else:
                    buff.append(c)
            else:
                buff.append(c)
        return "".join(buff)
    except MissingEscapedChar:
        raise MissingEscapedChar(vars_)
    except MissingClosingBrace:
        raise MissingClosingBrace(vars_)
    except BadSubstitution:
        raise BadSubstitution(vars_)


def expandvars(vars_, nounset=False):
    """Expand system variables Unix style.

    Params:
        vars_ (str): System variables to expand.
        nounset (bool): If True, enables strict parsing (similar to set -u / set -o nounset in bash).

    Returns:
        str: Expanded values.

    Example usage: ::

        from expandvars import expandvars

        print(expandvars("$PATH:$HOME/bin:${SOME_UNDEFINED_PATH:-/default/path}"))
        # /bin:/sbin:/usr/bin:/usr/sbin:/home/you/bin:/default/path

        # Or
        with open(somefile) as f:
            print(expandvars(f))
    """
    return expand(vars_, nounset=nounset)


class ModifierType:
    GET_DEFAULT = 1
    GET_OR_SET_DEFAULT = 2
    SUBSTITUTE = 3
    OFFSET = 4
    STRICT = 5
    LENGTH = 6


class State:
    READING_VAR = 1
    READING_MODIFIER_VAR = 2
    READING_MODIFIER = 3
    FINISHED_READING = -1


def _read_var(buff, var_symbol):
    name = []
    state = State.READING_VAR
    next_ = buff.peek()
    modifier, modifier_type = [], None
    bracedepth = 0
    indirect = False

    while state != State.FINISHED_READING:
        next_ = buff.peek()

        if next_ == _PeekableIterator.NOTHING:
            if state == State.READING_MODIFIER_VAR or state == State.READING_MODIFIER:
                raise MissingClosingBrace("".join(name))
            else:
                state = State.FINISHED_READING

        elif next_ == "{" and state == State.READING_VAR:
            next(buff)
            if buff.peek() == "!":
                indirect = True
                next(buff)
            elif buff.peek() == "#":
                modifier_type = ModifierType.LENGTH
                next(buff)
            state = State.READING_MODIFIER_VAR

        elif next_ == "}" and state == State.READING_MODIFIER_VAR:
            next(buff)
            state = State.FINISHED_READING

        elif (_valid_char(next_) or (next_ == var_symbol and not name)) and (
            state == State.READING_VAR or state == State.READING_MODIFIER_VAR
        ):
            name.append(next(buff))

        elif next_ == ":" and state == State.READING_MODIFIER_VAR:
            state = State.READING_MODIFIER
            next(buff)

        elif state == State.READING_MODIFIER_VAR or (
            state == State.READING_MODIFIER and modifier_type is None
        ):
            state = State.READING_MODIFIER
            modifier, modifier_type = [], None
            if next_ == "-":
                modifier_type = ModifierType.GET_DEFAULT
            elif next_ == "=":
                modifier_type = ModifierType.GET_OR_SET_DEFAULT
            elif next_ == "+":
                modifier_type = ModifierType.SUBSTITUTE
            elif next_ == "?":
                modifier_type = ModifierType.STRICT
            elif next_ == "}":
                modifier_type = ModifierType.OFFSET
                state = State.FINISHED_READING
            else:
                modifier_type = ModifierType.OFFSET
                modifier.append(next_)
            next(buff)

        elif state == State.READING_MODIFIER:
            c = next(buff)
            if c == "{":
                bracedepth += 1
                modifier.append(c)
            elif c == "}":
                if bracedepth == 0:
                    state = State.FINISHED_READING
                else:
                    modifier.append(c)
                    bracedepth -= 1
            else:
                modifier.append(c)

        elif state == State.READING_VAR:
            state = State.FINISHED_READING

        else:
            # This should never happen
            raise ExpandvarsException("unexpected state")  # pragma: no cover

    var = "".join(name)
    return var, modifier_type, modifier, indirect


def _expand_var(buff, nounset, environ, var_symbol):
    var, modifier_type, modifier, indirect = _read_var(buff, var_symbol=var_symbol)
    if not var:
        raise BadSubstitution("")

    val = getenv(var, indirect=indirect, environ=environ, var_symbol=var_symbol)
    modifier = expand("".join(modifier), environ=environ, var_symbol=var_symbol)

    if modifier_type == ModifierType.LENGTH:
        if modifier:
            raise BadSubstitution(var)
        modified = str(len(val or ""))
    elif modifier_type == ModifierType.GET_DEFAULT:
        modified = val if val else modifier

    elif modifier_type == ModifierType.GET_OR_SET_DEFAULT:
        modified = _modify_get_or_set_default(var, val, modifier, environ=environ)

    elif modifier_type == ModifierType.SUBSTITUTE:
        modified = modifier if val else ""

    elif modifier_type == ModifierType.OFFSET:
        modified = _modify_offset(var, val, modifier)

    elif modifier_type == ModifierType.STRICT:
        modified = _modify_strict(var, val, modifier, environ=environ)

    else:
        modified = val

    if modified is None:
        if nounset:
            recover_null = environ.get("EXPANDVARS_RECOVER_NULL", None)
            if recover_null is None:
                raise UnboundVariable(var)
            else:
                modified = recover_null
        else:
            modified = ""

    return modified


def _modify_get_or_set_default(var, val, modifier, environ):
    if val:
        return val
    else:
        environ[var] = modifier
        return modifier


def _modify_offset(var, val, modifier):
    if not modifier:
        raise BadSubstitution(var)

    split_ = modifier.split(":")

    if len(split_) == 1:
        offset_str, length_str = split_[0].strip(), None
    elif len(split_) == 2:
        offset_str, length_str = (s.strip() for s in split_)
    else:
        raise BadSubstitution(var)

    if _isint(offset_str):
        offset = int(offset_str)
    else:
        offset = 0

    if length_str is None:
        length = None
    elif not _isint(length_str):
        if not all(_valid_char(c) for c in length_str):
            raise OperandExpected(var, length_str)
        else:
            length = 0
    else:
        length = int(length_str)
        if length < 0:
            raise NegativeSubStringExpression(var, length_str)

    width = offset + length if length is not None else None
    return (val or "")[offset:width]


def _modify_strict(var, val, modifier, environ):
    if val:
        return val

    recover_null = environ.get("EXPANDVARS_RECOVER_NULL", None)
    if recover_null is not None:
        return recover_null
    raise ParameterNullOrNotSet(var, modifier if modifier else None)


def _valid_char(char):
    return char.isalnum() or char == "_"


def _isint(val):
    try:
        int(val)
        return True
    except ValueError:
        return False


class _PeekableIterator:
    """Peekable iterator."""

    NOTHING = object()

    def __init__(self, iterable):
        self.iterator = iter(iterable)
        self._next = self.NOTHING

    def __iter__(self):
        return self

    def __next__(self):
        if self._next is self.NOTHING:
            return next(self.iterator)
        else:
            next_ = self._next
            self._next = self.NOTHING
            return next_

    def peek(self):
        """Peek at the next item."""
        if self._next is self.NOTHING:
            self._next = next(self.iterator, self.NOTHING)
        return self._next



# -*- coding: utf-8 -*-

import importlib
from os import environ as env
from unittest.mock import patch

import expandvars


@patch.dict(env, {"EXPANDVARS_RECOVER_NULL": "foo", "BAR": "bar"}, clear=True)
def test_strict_parsing_recover_null():
    importlib.reload(expandvars)

    assert expandvars.expandvars("${FOO:?}:${BAR?}") == "foo:bar"
    assert expandvars.expandvars("${FOO:?custom err}:${BAR?custom err}") == "foo:bar"

    assert expandvars.expandvars("$FOO$BAR", nounset=True) == "foobar"
    assert expandvars.expandvars("${FOO}:${BAR}", nounset=True) == "foo:bar"
